(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udebug.ml,v 1.36 2014/02/10 14:36:40 fclement Exp $ *)

(* Debug/report/error message facilities. *)

module Cmd_opt = Ucommand_option;;

open Uaref;;

type purpose = | User | Library
and software_name = string
and tag = string
and 'a make_tag = 'a -> tag
;;

type short_key_prefix =
  | Automatic
  | Short of Upervasives.prefix

and key_prefixes =
  {
    long : Upervasives.prefix;
    short : short_key_prefix;
  }
;;

type 'a print = 'a -> unit
and 'a fun_name_print = Upervasives.function_name -> 'a print
;;

type printing_facilities =
  {
   report : Upervasives.message print;
   debug : Upervasives.message fun_name_print;
   warning : Upervasives.message fun_name_print;
   error : Upervasives.message fun_name_print;
   report_action : Upervasives.gen_message print;
   debug_action : Upervasives.gen_message fun_name_print;
   warning_action : Upervasives.gen_message fun_name_print;
   error_action : Upervasives.gen_message fun_name_print;
  }

and 'a dynamic_printing_facilities =
  {
   dyn_report : 'a -> Upervasives.message print;
   dyn_debug : 'a -> Upervasives.message fun_name_print;
   dyn_warning : 'a -> Upervasives.message fun_name_print;
   dyn_error : 'a -> Upervasives.message fun_name_print;
   dyn_report_action : 'a -> Upervasives.gen_message print;
   dyn_debug_action : 'a -> Upervasives.gen_message fun_name_print;
   dyn_warning_action : 'a -> Upervasives.gen_message fun_name_print;
   dyn_error_action : 'a -> Upervasives.gen_message fun_name_print;
  }
;;

(* Key prefixes. *)

let make_key_prefixes sname =
  { long = Ustring.long_key_of_string sname; short = Automatic; }
;;

(* Printing. *)

let display_message_on_stderr ob msg =
  Printf.bprintf ob "%s\n" msg;
  Ubuffer.flush_buffer_on_stderr ob
;;

(* Messages with dynamic tags. *)

let make_printing_facilities_with_dynamic_tags
    purpose sname { long = long_kp; short = short_kp; } make_tag =
  
  (* Report and debug flags. *)
  
  let (report_flag, debug_flag) = make_embedded_flags () in
  
  let add_flag kind k comment set_flag add =
    let short_key =
      match short_kp with
      | Automatic -> Cmd_opt.with_auto_short_key
      | Short skp ->
          Cmd_opt.with_short_key (Printf.sprintf "-%s%c" skp k) in
    add
      (short_key
         (Cmd_opt.with_auto_gnu_key
            (Cmd_opt.of_regular
               { Cmd_opt.key = Printf.sprintf "%s-%s" long_kp kind;
                 spec = Arg.Unit set_flag;
                 doc =
                   Printf.sprintf
                     "\n\tTurn %s %s mode on. Default is off.%s"
                     sname kind comment; }))) in
  
  let add_report_flag () =
    let report_add =
      match purpose with
      | User -> Cmd_opt.report_user_add
      | Library -> Cmd_opt.report_library_add in
    add_flag "report" 'k' ""
      report_flag.set_flag report_add in
  add_report_flag ();

  let add_debug_flag () =
    let debug_add =
      match purpose with
      | User -> Cmd_opt.debug_user_add
      | Library -> Cmd_opt.debug_library_add in
    add_flag "debug" 'd'
      (Printf.sprintf "\n\tImplies %s report mode." sname)
      debug_flag.set_flag debug_add in
  add_debug_flag ();

  (* Printing functions. *)
  
  let ob = Buffer.create 1024 in
  
  let report_action x action =
    if report_flag.get_flag () then
      let tag = make_tag x in
      let msg = Upervasives.report_message tag (action ()) in
      display_message_on_stderr ob msg in
  
  let report x msg = report_action x (fun () -> msg) in
  
  let gen_action get_flag gen_message fun_name x action =
    if get_flag () then
      let tag = make_tag x in
      let msg = gen_message tag fun_name (action ()) in
      display_message_on_stderr ob msg in
  
  let debug_action x fun_name action =
    let get_flag = debug_flag.get_flag in
    gen_action get_flag Upervasives.debug_message fun_name x action in
  
  let debug x fun_name msg = debug_action x fun_name (fun () -> msg) in
  
  let warning_action x fun_name action =
    let get_flag () = true in
    gen_action get_flag Upervasives.warning_message fun_name x action in
  
  let warning x fun_name msg = warning_action x fun_name (fun () -> msg) in
  
  let error_action x fun_name action =
    let get_flag () = true in
    gen_action get_flag Upervasives.error_message fun_name x action in
  
  let error x fun_name msg = error_action x fun_name (fun () -> msg) in
  
  {
   dyn_report = report;
   dyn_debug = debug;
   dyn_warning = warning;
   dyn_error = error;
   dyn_report_action = report_action;
   dyn_debug_action = debug_action;
   dyn_warning_action = warning_action;
   dyn_error_action = error_action;
  }
;;

let gen_make_printing_facilities purpose sname kp make_tag =
  let
      {
       dyn_report = report;
       dyn_debug = debug;
       dyn_warning = warning;
       dyn_error = error;
       dyn_report_action = report_action;
       dyn_debug_action = debug_action;
       dyn_warning_action = warning_action;
       dyn_error_action = error_action;
      } =
    make_printing_facilities_with_dynamic_tags purpose sname kp make_tag in
  {
   report = report ();
   debug = debug ();
   warning = warning ();
   error = error ();
   report_action = report_action ();
   debug_action = debug_action ();
   warning_action = warning_action ();
   error_action = error_action ();
  }
;;

let make_printing_facilities purpose sname kp =
  let make_tag () = "" in
  gen_make_printing_facilities purpose sname kp make_tag
;;

let make_printing_facilities_with_static_tag purpose sname kp =
  let make_tag () = Printf.sprintf "[%s]" sname in
  gen_make_printing_facilities purpose sname kp make_tag
;;

(* Specializing printed messages. *)

type specialized_printing_facilities =
  {
   spf_report : Upervasives.message print;
   spf_debug : Upervasives.message print;
   spf_warning : Upervasives.message print;
   spf_error : Upervasives.message print;
   spf_report_action : Upervasives.gen_message print;
   spf_debug_action : Upervasives.gen_message print;
   spf_warning_action : Upervasives.gen_message print;
   spf_error_action : Upervasives.gen_message print;
  }
;;

let set_function_name fun_name =
  let set_fun_name print = print fun_name in
  fun pf ->
    {
     spf_report = pf.report;
     spf_debug = set_fun_name pf.debug;
     spf_warning = set_fun_name pf.warning;
     spf_error = set_fun_name pf.error;
     spf_report_action = pf.report_action;
     spf_debug_action = set_fun_name pf.debug_action;
     spf_warning_action = set_fun_name pf.warning_action;
     spf_error_action = set_fun_name pf.error_action;
    }
;;

let prefix_messages pref =
  let lift print msg = print (Upervasives.prefix pref msg) in
  let lift_action print_action action =
    print_action (fun () -> Upervasives.prefix pref (action ())) in
  fun spf ->
    {
     spf_report = lift spf.spf_report;
     spf_debug = lift spf.spf_debug;
     spf_warning = lift spf.spf_warning;
     spf_error = lift spf.spf_error;
     spf_report_action = lift_action spf.spf_report_action;
     spf_debug_action = lift_action spf.spf_debug_action;
     spf_warning_action = lift_action spf.spf_warning_action;
     spf_error_action = lift_action spf.spf_error_action;
    }

type 'a specialized_dynamic_printing_facilities =
  {
   sdpf_report : 'a -> Upervasives.message print;
   sdpf_debug : 'a -> Upervasives.message print;
   sdpf_warning : 'a -> Upervasives.message print;
   sdpf_error : 'a -> Upervasives.message print;
   sdpf_report_action : 'a -> Upervasives.gen_message print;
   sdpf_debug_action : 'a -> Upervasives.gen_message print;
   sdpf_warning_action : 'a -> Upervasives.gen_message print;
   sdpf_error_action : 'a -> Upervasives.gen_message print;
  }
;;

let set_function_name_dyn fun_name =
  let set_fun_name dyn_print x = dyn_print x fun_name in
  fun dpf ->
    {
     sdpf_report = dpf.dyn_report;
     sdpf_debug = set_fun_name dpf.dyn_debug;
     sdpf_warning = set_fun_name dpf.dyn_warning;
     sdpf_error = set_fun_name dpf.dyn_error;
     sdpf_report_action = dpf.dyn_report_action;
     sdpf_debug_action = set_fun_name dpf.dyn_debug_action;
     sdpf_warning_action = set_fun_name dpf.dyn_warning_action;
     sdpf_error_action = set_fun_name dpf.dyn_error_action;
    }
;;

let prefix_messages_dyn pref =
  let lift dyn_print x msg = dyn_print x (Upervasives.prefix pref msg) in
  let lift_action dyn_print_action x action =
    dyn_print_action x (fun () -> Upervasives.prefix pref (action ())) in
  fun sdpf ->
    {
     sdpf_report = lift sdpf.sdpf_report;
     sdpf_debug = lift sdpf.sdpf_debug;
     sdpf_warning = lift sdpf.sdpf_warning;
     sdpf_error = lift sdpf.sdpf_error;
     sdpf_report_action = lift_action sdpf.sdpf_report_action;
     sdpf_debug_action = lift_action sdpf.sdpf_debug_action;
     sdpf_warning_action = lift_action sdpf.sdpf_warning_action;
     sdpf_error_action = lift_action sdpf.sdpf_error_action;
    }
;;
