(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udebug.mli,v 1.30 2014/02/10 14:36:40 fclement Exp $ *)

(** {3 Report/debug/warning/error option flags and printing facilities}
 
 This module provides functionalities related to the set up of specific
 reporting and debugging option flags (either marked for [User] of [Library]
 purpose) to be used on the command line, and specific printing facilities for
 reporting/monitoring, debugging, and printing warning and error messages.
 Messages are printed on stderr.
 
 This module is based on modules {!Upervasives}, and {!Ucommand_option}
 (from the OCaml user's additional library Ulib). *)

(** {6 Messages with no tags} *)

type purpose = | User | Library
and software_name = string
;;
(** Basic types for [Udebug]. *)

type short_key_prefix =
  | Automatic
  | Short of Upervasives.prefix

and key_prefixes =
  {
   long : Upervasives.prefix;
   short : short_key_prefix;
  }
;;
(** The type of key prefixes.
 Long and short prefixes to build dedicated reporting and debugging option
 flags to be used on the command line.
 When the [short] field equals [Automatic], the short key prefix is
 eventually computed from the initials of the words of the [long] field
 (where the separator is the dash character ['-']). *)

type 'a print = 'a -> unit
and 'a fun_name_print = Upervasives.function_name -> 'a print
;;

type printing_facilities =
  {
   report : Upervasives.message print;
   debug : Upervasives.message fun_name_print;
   warning : Upervasives.message fun_name_print;
   error : Upervasives.message fun_name_print;
   report_action : Upervasives.gen_message print;
   debug_action : Upervasives.gen_message fun_name_print;
   warning_action : Upervasives.gen_message fun_name_print;
   error_action : Upervasives.gen_message fun_name_print;
  }
;;
(** The type of printing facilities.
 Prefixes following the specification described above are automatically added.
 Field [report] is a printing function taking the message to be prefixed as
 argument.
 Fields [debug], [warning], and [error] are similar to [report] but they take
 the function name as an additional first argument.
 Fields [report_action], [debug_action], [warning_action], and [error_action]
 are similar to [report], [debug], [warning], and [error] with their message
 argument replaced by a message generator argument. *)

val make_printing_facilities :
  purpose -> software_name -> key_prefixes -> printing_facilities
;;
(** Printing facilities construction function with no tag prefix.
 Reporting and debugging option flags to be used on the command line are
 automatically computed from the key prefixes. *)

val make_key_prefixes : software_name -> key_prefixes;;
(** Key prefixes construction function.
 The long key prefix is a stripped version of the software name (lowercase,
 all consecutive non alphanumeric characters are replaced by a single dash
 (['-']) character, and characters from the ISO Latin-1 (8859-1) character
 set are mapped to the US-ASCII character set).
 The short key prefix option is [Automatic] (see type {!key_prefixes}). *)

(** {6 Messages with static tags} *)

val make_printing_facilities_with_static_tag :
  purpose -> software_name -> key_prefixes -> printing_facilities
;;
(** Printing facilities construction function with static tag prefix.
 It is similar to the printing facilities construction function
 {!make_printing_facilities} except that a static tag is prefixed to each
 printed message. This tag is the software name between square brackets
 characters (['['], [']']). *)

(** {6 Messages with dynamic tags} *)

type tag = string
and 'a make_tag = 'a -> tag
;;
(** Dynamic tags are functions returning a string. *)

type 'a dynamic_printing_facilities =
  {
   dyn_report : 'a -> Upervasives.message print;
   dyn_debug : 'a -> Upervasives.message fun_name_print;
   dyn_warning : 'a -> Upervasives.message fun_name_print;
   dyn_error : 'a -> Upervasives.message fun_name_print;
   dyn_report_action : 'a -> Upervasives.gen_message print;
   dyn_debug_action : 'a -> Upervasives.gen_message fun_name_print;
   dyn_warning_action : 'a -> Upervasives.gen_message fun_name_print;
   dyn_error_action : 'a -> Upervasives.gen_message fun_name_print;
  }
;;
(** The type of dynamic printing facilities.
 The fields are the similar to those of the type {!printing_facilities} except
 that they have an additional first argument to compute the dynamic tag. *)

val make_printing_facilities_with_dynamic_tags :
  purpose -> software_name -> key_prefixes -> 'a make_tag ->
  'a dynamic_printing_facilities
;;
(** Printing facilities construction function with dynamic tag prefixes.
 It is similar to the (static) printing facilities construction function
 {!make_printing_facilities_with_static_tag} except that the tag generator
 is given as an additional last argument. *)

(** {6 Specializing printed messages} *)

type specialized_printing_facilities =
  {
   spf_report : Upervasives.message print;
   spf_debug : Upervasives.message print;
   spf_warning : Upervasives.message print;
   spf_error : Upervasives.message print;
   spf_report_action : Upervasives.gen_message print;
   spf_debug_action : Upervasives.gen_message print;
   spf_warning_action : Upervasives.gen_message print;
   spf_error_action : Upervasives.gen_message print;
  }
;;

val set_function_name :
  Upervasives.function_name ->
  printing_facilities -> specialized_printing_facilities
;;

val prefix_messages :
  Upervasives.prefix ->
  specialized_printing_facilities -> specialized_printing_facilities
;;

type 'a specialized_dynamic_printing_facilities =
  {
   sdpf_report : 'a -> Upervasives.message print;
   sdpf_debug : 'a -> Upervasives.message print;
   sdpf_warning : 'a -> Upervasives.message print;
   sdpf_error : 'a -> Upervasives.message print;
   sdpf_report_action : 'a -> Upervasives.gen_message print;
   sdpf_debug_action : 'a -> Upervasives.gen_message print;
   sdpf_warning_action : 'a -> Upervasives.gen_message print;
   sdpf_error_action : 'a -> Upervasives.gen_message print;
  }
;;

val set_function_name_dyn :
  Upervasives.function_name ->
  'a dynamic_printing_facilities ->
  'a specialized_dynamic_printing_facilities
;;

val prefix_messages_dyn :
  Upervasives.prefix ->
  'a specialized_dynamic_printing_facilities ->
  'a specialized_dynamic_printing_facilities
;;
