(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: upervasives.ml,v 1.36 2014/02/10 14:36:40 fclement Exp $ *)

(** Extensions to the initially opened module. *)

(* Report/debug/warning/error messages. *)

type module_name = string
and function_name = string
and tag = string
and prefix = string
and message = string
and gen_message = unit -> message
;;

let combine sep s1 s2 =
  if String.length s1 = 0 then s2 else
  if String.length s2 = 0 then s1 else
  Printf.sprintf "%s%s%s" s1 sep s2
;;

let prefix = combine " "
and qualify = combine "."
and quote = combine ": "
;;

let prefix_and_quote tag fun_name msg = prefix tag (quote fun_name msg);;
let prefix_quote_and_prefix ann tag fun_name =
  prefix_and_quote tag (prefix ann fun_name)
;;

let report_message = prefix
and debug_message = prefix_and_quote
and warning_message = prefix_quote_and_prefix "Warning in"
and error_message = prefix_quote_and_prefix "Fatal error in"
;;

(* Exceptions. *)

let invalid_arg fun_name msg =
  Pervasives.invalid_arg (error_message "" fun_name msg)
;;

let failwith fun_name msg =
  Pervasives.failwith (error_message "" fun_name msg)
;;

let internal_error fun_name =
  failwith fun_name "Internal error, please report."
;;

let compose g f x = g (f x);;

let invalid_arg_qual qual = compose invalid_arg qual
and failwith_qual qual = compose failwith qual
and internal_error_qual qual = compose internal_error qual
;;

(* Functional helpers. *)

external id : 'a -> 'a = "%identity";;

let compose g f x = g (f x);;

let proj2_1 x _ = x;;
let proj2_2 _ y = y;;

let option f d = function
  | Some x -> f x
  | None -> d
;;

let lifto f = option (fun x -> Some (f x)) None;;

(* Integer arithmetic. *)

type index = Unatural.t;;

let index_of_int = Unatural.of_int;;
let index_to_int = Unatural.to_int;;

type length = Unatural.t;;

let length_of_int = Unatural.of_int;;
let length_to_int = Unatural.to_int;;

(* Floating-point arithmetic. *)

let round_pos x = int_of_float (x +. 0.5);;
let round_neg x = int_of_float (x -. 0.5);;
let round x = if x >= 0. then round_pos x else round_neg x;;

(* Conversion. *)

let char_of_int64 i = char_of_int (Int64.to_int (Int64.logand i 0xffL));;

let string_init n f =
  if n = 0 then "" else
  let res =
    try String.make n (f 0) with
    | Invalid_argument _ -> Pervasives.invalid_arg "string_init" in
  for i = 1 to pred n do
    res.[i] <- f i;
  done;
  res
;;

let bit_string_of_float f =
  let i = Int64.bits_of_float f in
  string_init 8 (fun j -> char_of_int64 (Int64.shift_right i (8*j)))
;;

let int64_of_char c = Int64.of_int (int_of_char c);;

let float_of_bit_string s =
  assert (String.length s = 8);
  let rec loop accu j =
    if j < 0 then accu else
    loop (Int64.add (Int64.shift_left accu 8) (int64_of_char s.[j])) (j - 1) in
  Int64.float_of_bits (loop (int64_of_char s.[7]) 6)
;;

(* Input/output. *)

let pr_unit pr_string () = pr_string "()";;

let print_unit = pr_unit print_string;;

let prerr_unit = pr_unit prerr_string;;

let pr_bool pr_string b =
  if b then pr_string "true" else pr_string "false"
;;

let print_bool = pr_bool print_string;;

let prerr_bool = pr_bool prerr_string;;
  

let open_append fname =
  let mode = [Open_wronly; Open_append; Open_creat; Open_text;] in
  let perm = 0o644 in
  open_out_gen mode perm fname
;;

let open_append_bin fname =
  let mode = [Open_wronly; Open_append; Open_creat; Open_binary;] in
  let perm = 0o644 in
  open_out_gen mode perm fname
;;
