(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ustring.ml,v 1.22 2014/02/09 17:27:28 fclement Exp $ *)

(* More string operations. *)

let s_qual = Upervasives.qualify "Ustring";;

let s_invalid_arg_no_msg fun_name =
  Upervasives.invalid_arg_qual s_qual fun_name ""
;;
let s_invalid_arg = s_invalid_arg_no_msg
and s_failwith = Upervasives.failwith_qual s_qual
;;

type length = Upervasives.length;;

let length s = Upervasives.length_of_int (String.length s);;

let create n = String.create (Upervasives.length_to_int n);;

let make n c = String.make (Upervasives.length_to_int n) c;;

let init n f =
  if n = Unatural._0_ then "" else
  let res =
    try make n (f 0) with
    | Invalid_argument _ -> s_invalid_arg "init" in
  for i = 1 to pred (Upervasives.length_to_int n) do
    res.[i] <- f i;
  done;
  res
;;

let sub s ofs len =
  if ofs < 0 ||
    ofs + Upervasives.length_to_int len >
      Upervasives.length_to_int (length s) then s_invalid_arg "sub" else
  let r = create len in
  String.unsafe_blit s ofs r 0 (Upervasives.length_to_int len);
  r
;;

let fill s ofs len c =
  if ofs < 0 ||
    ofs + Upervasives.length_to_int len >
      Upervasives.length_to_int (length s) then s_invalid_arg "fill" else
  String.unsafe_fill s ofs (Upervasives.length_to_int len) c
;;

let blit s1 ofs1 s2 ofs2 len =
  if ofs1 < 0 ||
    ofs1 + Upervasives.length_to_int len >
      Upervasives.length_to_int (length s1) ||
    ofs2 < 0 ||
    ofs2 + Upervasives.length_to_int len >
      Upervasives.length_to_int (length s2) then s_invalid_arg "blit" else
  String.unsafe_blit s1 ofs1 s2 ofs2 (Upervasives.length_to_int len)
;;

let replace s1 ofs1 len1 s2 ofs2 len2 =
  let l1 = length s1 in
  let l2 = length s2 in
  if ofs1 < 0 ||
    ofs1 + Upervasives.length_to_int len1 >
      Upervasives.length_to_int l1 ||
    ofs2 < 0 ||
    ofs2 + Upervasives.length_to_int len2 >
      Upervasives.length_to_int l2 then s_invalid_arg "replace" else
  let len3 = Unatural.sub (Unatural.add l1 len2) len1 in
  let s3 = create len3 in
  blit s1 0 s3 0 (Upervasives.length_of_int ofs1);
  blit s2 ofs2 s3 ofs1 len2;
  blit
    s1 (ofs1 + Upervasives.length_to_int len1)
    s3 (ofs1 + Upervasives.length_to_int len2)
    (Unatural.sub l1 (Unatural.add (Upervasives.length_of_int ofs1) len1));
  s3
;;

let rec fold_left f accu s =
  let len = String.length s in
  if len = 0 then accu else
  fold_left f (f accu s.[0]) (String.sub s 1 (len - 1))
;;

let indices s c =
  let rec loop accu i =
    try
      let j = String.rindex_from s i c in
      loop (j :: accu) (j - 1)
    with
    | Not_found -> accu in
  loop [] (String.length s - 1)
;;

type prefix = string
and suffix = string
;;

let chop_prefix pref s =
  let len_pref = String.length pref in
  let len = String.length s - len_pref in
  if len < 0 then s else
  if String.sub s 0 len_pref <> pref then s else
  String.sub s len_pref len
;;

let chop_suffix suff s =
  let len_suff = String.length suff in
  let len = String.length s - len_suff in
  if len < 0 then s else
  if String.sub s len len_suff <> suff then s else
  String.sub s 0 len
;;

let trim_edges p s =
  let len = String.length s in
  let rec loop i j =
    if p s.[i] then loop (i + 1) j else
    if p s.[j - 1] then loop i (j - 1) else
    (i, j) in
  let (i, j) = loop 0 len in
  if i = 0 && j = len then s else
  if j > i then String.sub s i (j - i) else
  ""
;;

let trim_consecutive_chars skip_it s1 =
  let len1 = String.length s1 in
  if len1 = 0 then s1 else
  let s2 = String.copy s1 in
  let rec loop i1 i2 prev =
    if i1 >= len1 then i2 else
    if skip_it s1 i1 prev
    then loop (i1 + 1) i2 prev
    else begin
      s2.[i2] <- s1.[i1];
      loop (i1 + 1) (i2 + 1) s1.[i1]
    end in
  let len2 = loop 1 1 s1.[0] in
  if len2 = len1 then s1 else
  String.sub s2 0 len2
;;

let trim_consecutive p =
  trim_consecutive_chars
    (fun s1 i1 prev -> s1.[i1] = prev && p s1.[i1])
;;

let trim_doubles =
  trim_consecutive_chars (fun s1 i1 prev -> s1.[i1] = prev)
;;

let iso_latin1_to_us_ascii string_of_unextended_char s =
  let buf = Buffer.create 1024 in
  let buffer_add_iso_latin1_char c =
    try Buffer.add_char buf (Uchar.iso_latin1_to_us_ascii c) with
    | Uchar.Unextended_char c ->
      Buffer.add_string buf (string_of_unextended_char c)
    | Uchar.Digram (c1, c2) ->
      Buffer.add_char buf c1;
      Buffer.add_char buf c2 in
  String.iter buffer_add_iso_latin1_char s;
  Buffer.contents buf
;;

(* Keys. *)

type key = string
and long_key = key
and short_key = key
;;

let mask_non_alphanumeric mask c =
  if Uchar.is_alphanumeric c then c else mask
;;

let long_key_of_string s =
  let mask = '-' in
  let is_mask c = c = mask in
  let s' =
    trim_edges is_mask
      (trim_consecutive is_mask
        (String.lowercase
          (iso_latin1_to_us_ascii (fun _ -> "")
            (String.map (mask_non_alphanumeric mask) s)))) in
  if String.length s' = 0 then
    s_failwith "long_key_of_string"
      (Printf.sprintf "Empty long key prefix from string %S" s) else
  "-" ^ s'
;;

let is_dash c = c = '-';;

let short_key_of_long_key lk =
  let len = String.length lk in
  if len = 0 then
    invalid_arg "Empty long key" else
  if not (is_dash lk.[0]) then
    invalid_arg (Printf.sprintf "%S does not start with a dash" lk) else
  if is_dash lk.[len - 1] then
    invalid_arg (Printf.sprintf "%S ends with a dash" lk) else
  let lk = trim_consecutive is_dash lk in
  match indices lk '-' with
  | [] -> assert false
  | is ->
    let len = Upervasives.length_of_int (List.length is) in
    let sk = create len in
    List.iteri (fun j i -> sk.[j] <- lk.[i + 1]) is;
    "-" ^ sk
;;
